"""Settings management."""

import logging
import os
from collections.abc import Sequence
from logging.config import dictConfig
from typing import Any, Generic, Optional, TypeVar, cast

import yaml
from pydantic import BaseModel
from pydantic_settings import (
    BaseSettings as PydanticBaseSettings,
    InitSettingsSource,
    PydanticBaseSettingsSource,
    SettingsConfigDict as PydanticSettingsConfigDict,
)

from .constants import Verbosity

try:  # pragma: no cover
    import sentry_sdk
except ImportError:
    sentry_sdk = None

_LOGGER = logging.getLogger(__name__)


class YamlConfigSettingsSource(InitSettingsSource):
    """A source class that loads variables from a yaml file."""

    def __init__(self, settings_cls: type[PydanticBaseSettings]):
        self.data = {}
        config_file = settings_cls.model_config.get("config_file")
        if not config_file and settings_cls.model_config.get("config_environ") is not None:
            config_file = os.environ.get(cast(str, settings_cls.model_config.get("config_environ")))
        if not config_file:
            for name in cast(Sequence[str], settings_cls.model_config.get("config_files", [])):
                full_name = os.path.expanduser(name)
                if os.path.isfile(full_name):
                    config_file = full_name
                    break
        if config_file:
            with open(os.path.expanduser(cast(str, config_file))) as file:
                self.data = cast(dict[str, Any], yaml.safe_load(file))
        return super().__init__(settings_cls, self.data)


class SettingsConfigDict(PydanticSettingsConfigDict, total=False):
    """Customized settings config.

    Class attributes:
        config_file: User supplied config file.
        config_environ: Name of the environment variable with the custom configuration file.
        config_files: A sequence of default config files in YAML format.
                      The first file found is used as a configuration file.
                      User home directory placeholders are expanded.

    Order of preference is: config_file, config_environ, config_files.
    """

    config_file: Optional[str]
    config_environ: Optional[str]
    config_files: Sequence[str]


class SentrySettings(BaseModel):
    """Sentry settings."""

    dsn: str
    environment: Optional[str] = None
    release: Optional[str] = None
    ca_certs: Optional[str] = None


class BaseSettings(PydanticBaseSettings):
    """Application settings.

    When you inherit from this class, you should specify `env_prefix`, `config_files`
    and `config_environ` in `model_config`.
    The `model_config` attribute should be an instance of setapp.SettingsConfigDict.
    """

    logging: dict = {}
    sentry: Optional[SentrySettings] = None

    model_config = SettingsConfigDict(
        extra="ignore",
        case_sensitive=False,
        env_nested_delimiter="__",
        config_file=None,
        config_files=(),
        config_environ=None,
    )

    def __init__(self, config_file: Optional[str] = None, **kwargs: Any) -> None:
        """Initialize settings.

        Arguments:
            config_file: Name of the custom configuration file. Has precedence over other config file sources.
            **kwargs: Custom settings.
        """
        self.model_config["config_file"] = config_file
        super().__init__(**kwargs)

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[PydanticBaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        """Include config file in configuration sources."""
        # first item has the highest priority
        return (
            init_settings,
            env_settings,
            # dotenv settings are purposefully left out
            YamlConfigSettingsSource(settings_cls),
            file_secret_settings,
        )


S = TypeVar("S", bound=BaseSettings)


class BaseApplication(Generic[S]):
    """Represent application settings and basic setup.

    Attributes:
        settings_cls: The default settings class for the application.
                      Unset by default.
    """

    settings_cls: type[S] = None  # type: ignore[assignment]
    verbosity: Verbosity = Verbosity.INFO

    def __init__(self, settings_cls: Optional[type[S]] = None) -> None:
        self.settings_cls = settings_cls or self.settings_cls
        if self.settings_cls is None:
            raise ValueError("Settings class unset")
        self._settings: Optional[S] = None

    @property
    def settings(self) -> S:
        """Return application settings."""
        if not self.is_loaded:
            raise RuntimeError("Application is not loaded")
        return cast(S, self._settings)

    @settings.setter
    def settings(self, value: S) -> None:
        """Set new application settings."""
        self._settings = value

    @property
    def is_loaded(self) -> bool:
        """Return whether application is loaded."""
        return self._settings is not None

    def load(self, config_file: Optional[str] = None, *, settings_kwargs: Optional[dict[str, Any]] = None) -> None:
        """Load application."""
        if self.is_loaded:
            raise RuntimeError("Application already loaded")
        self._settings = self.settings_cls(config_file=config_file, **settings_kwargs or {})
        self._setup_logs()

    def _setup_logs(self) -> None:
        """Set up logging systems."""
        # Apply logging configuration
        if self.settings.logging:
            dictConfig(self.settings.logging)

        # Set up Sentry
        if self.settings.sentry:
            if sentry_sdk is None:
                _LOGGER.warning("Sentry set up, but sentry-sdk not available.")
            else:
                sentry_sdk.init(**self.settings.sentry.model_dump())

    def _del(self, name: str) -> None:
        """Delete attribute if it exists.

        Utility function to invalidate `cached_property`.
        Designed to be called from `clear_cache` method.
        """
        try:
            delattr(self, name)
        except AttributeError:
            pass

    def clear_cache(self) -> None:
        """Clear all cached attributes."""
