import asyncio
import io
import os
from unittest import TestCase, TextTestResult
from unittest.mock import sentinel
from unittest.runner import _WritelnDecorator  # type: ignore[attr-defined, unused-ignore]

from setapp.settings import BaseApplication, BaseSettings, SentrySettings
from setapp.utils import override_environ, override_settings


class OverrideEnvironTest(TestCase):
    def setUp(self):
        self.orig_environ = os.environ.copy()

    def tearDown(self):
        os.environ.clear()
        os.environ.update(self.orig_environ)

    def test_override(self):
        os.environ["RIMMER"] = "ARNOLD"
        with override_environ(KRYTEN="2X4B"):
            self.assertEqual(os.environ["KRYTEN"], "2X4B")
            self.assertNotIn("RIMMER", os.environ)

        self.assertEqual(os.environ["RIMMER"], "ARNOLD")
        self.assertNotIn("KRYTEN", os.environ)

    def test_override_error(self):
        os.environ["RIMMER"] = "ARNOLD"
        with self.assertRaisesRegex(Exception, "Gazpacho!"):
            with override_environ(KRYTEN="2X4B"):
                raise Exception("Gazpacho!")

        self.assertEqual(os.environ["RIMMER"], "ARNOLD")
        self.assertNotIn("KRYTEN", os.environ)


class OverrideSettingsTest(TestCase):
    def test_context_manager(self):
        application = BaseApplication(BaseSettings)
        application.load()
        application.settings.sentry = SentrySettings(dsn="example.org")

        with override_settings(application, logging={"secret": "Gazpacho!"}):
            self.assertEqual(application.settings.logging, {"secret": "Gazpacho!"})
            self.assertEqual(application.settings.sentry.dsn, "example.org")

        # Check settings are back to previous state
        self.assertEqual(application.settings.logging, {})
        self.assertEqual(application.settings.sentry.dsn, "example.org")

    def test_context_manager_error(self):
        application = BaseApplication(BaseSettings)
        application.load()
        application.settings.sentry = SentrySettings(dsn="example.org")

        with self.assertRaisesRegex(Exception, "Quagaars!"):
            with override_settings(application, logging={"secret": "Gazpacho!"}):
                raise Exception("Quagaars!")

        # Check settings are back to previous state
        self.assertEqual(application.settings.logging, {})
        self.assertEqual(application.settings.sentry.dsn, "example.org")

    def test_decorate_any_class(self):
        with self.assertRaisesRegex(TypeError, "Can only decorate unittest.TestCase or function."):

            @override_settings(sentinel.application)
            class NotATestCase:
                pass

    def test_decorate_non_callable(self):
        with self.assertRaisesRegex(TypeError, "Can only decorate unittest.TestCase or function."):
            override_settings(sentinel.application)(sentinel.not_callable)

    def test_decorate_test_case(self):
        application = BaseApplication(BaseSettings)
        application.load()
        application.settings.sentry = SentrySettings(dsn="example.org")

        @override_settings(application, logging={"secret": "Gazpacho!"})
        class DecoratorTestCase(TestCase):
            def test_override(self):
                self.assertEqual(application.settings.logging, {"secret": "Gazpacho!"})
                self.assertEqual(application.settings.sentry.dsn, "example.org")  # type: ignore[union-attr]

        # Run the inner test
        buff = io.StringIO()
        result = TextTestResult(_WritelnDecorator(buff), True, 1)
        DecoratorTestCase("test_override").run(result)

        # Check the inner test passed
        result.printErrors()
        self.assertTrue(result.wasSuccessful(), msg=buff.getvalue())

        # Check settings are back to previous state
        self.assertEqual(application.settings.logging, {})
        self.assertEqual(application.settings.sentry.dsn, "example.org")

    def test_decorate_test_case_set_up_error(self):
        # Test override on test case with an error in setUp.
        application = BaseApplication(BaseSettings)
        application.load()
        application.settings.sentry = SentrySettings(dsn="example.org")

        @override_settings(application, logging={"secret": "Gazpacho!"})
        class DecoratorTestCase(TestCase):
            def setUp(self):
                raise Exception("Quagaars!")

            def test_override(self):  # pragma: no cover
                self.fail("Should not be called.")

        # Run the inner test
        buff = io.StringIO()
        result = TextTestResult(_WritelnDecorator(buff), True, 1)
        DecoratorTestCase("test_override").run(result)

        # Check the inner test failed
        result.printErrors()
        self.assertFalse(result.wasSuccessful(), msg=buff.getvalue())

        # Check settings are back to previous state
        self.assertEqual(application.settings.logging, {})
        self.assertEqual(application.settings.sentry.dsn, "example.org")

    def test_decorate_function(self):
        application = BaseApplication(BaseSettings)
        application.load()
        application.settings.sentry = SentrySettings(dsn="example.org")

        @override_settings(application, logging={"secret": "Gazpacho!"})
        def test_function(*args, **kwargs):
            self.assertEqual(application.settings.logging, {"secret": "Gazpacho!"})
            self.assertEqual(application.settings.sentry.dsn, "example.org")  # type: ignore[union-attr]
            self.assertEqual(args, (sentinel.arg,))
            self.assertEqual(kwargs, {"key": sentinel.value})
            return sentinel.result

        result = test_function(sentinel.arg, key=sentinel.value)

        self.assertEqual(result, sentinel.result)

        # Check settings are back to previous state
        self.assertEqual(application.settings.logging, {})
        self.assertEqual(application.settings.sentry.dsn, "example.org")

    def test_decorate_async_function(self):
        application = BaseApplication(BaseSettings)
        application.load()
        application.settings.sentry = SentrySettings(dsn="example.org")

        @override_settings(application, logging={"secret": "Gazpacho!"})
        async def test_function(*args, **kwargs):
            self.assertEqual(application.settings.logging, {"secret": "Gazpacho!"})
            self.assertEqual(application.settings.sentry.dsn, "example.org")  # type: ignore[union-attr]
            self.assertEqual(args, (sentinel.arg,))
            self.assertEqual(kwargs, {"key": sentinel.value})
            return sentinel.result

        result = asyncio.run(test_function(sentinel.arg, key=sentinel.value))

        self.assertEqual(result, sentinel.result)

        # Check settings are back to previous state
        self.assertEqual(application.settings.logging, {})
        self.assertEqual(application.settings.sentry.dsn, "example.org")
