"""Various utilities, mainly for tests."""

import asyncio
import os
from collections.abc import Generator
from contextlib import contextmanager
from functools import wraps
from types import TracebackType
from typing import Any, Callable, Optional, TypeVar, cast
from unittest import TestCase
from unittest.mock import Mock, patch

from .settings import BaseApplication, BaseSettings


@contextmanager
def override_environ(**kwargs: str) -> Generator[Mock, None, None]:
    """Clear the environment and set only provided variables."""
    with patch.dict(os.environ) as mock:
        os.environ.clear()
        os.environ.update(kwargs)
        yield mock


C = TypeVar("C", bound=Callable)
D = TypeVar("D", type[TestCase], Callable)


class override_settings:
    """Decorator and context manager to temporarily override settings."""

    def __init__(self, application: BaseApplication, /, **overrides: Any):
        self.application = application
        self.overrides = overrides
        self.old_settings: Optional[BaseSettings] = None

    def enable(self) -> None:
        """Start settings override."""
        self.old_settings = self.application.settings

        # Create a new instance of settings with overrides
        new_settings = dict(self.application.settings.model_dump(), **self.overrides)
        self.application.settings = self.application.settings_cls(**new_settings)

        # Clear caches
        self.application.clear_cache()

    def disable(self) -> None:
        """Stop settings override."""
        self.application.settings = cast(BaseSettings, self.old_settings)
        self.application.clear_cache()

    def __enter__(self) -> None:
        self.enable()

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.disable()

    def _decorate_test_case(self, decorated: type[TestCase]) -> type[TestCase]:
        original_setup = decorated.setUp
        original_teardown = decorated.tearDown

        outer_self = self

        def setUp(self: TestCase) -> None:
            outer_self.enable()
            try:
                original_setup(self)
            except Exception:
                outer_self.disable()
                raise

        def tearDown(self: TestCase) -> None:
            original_teardown(self)
            outer_self.disable()

        decorated.setUp = setUp  # type: ignore[method-assign]
        decorated.tearDown = tearDown  # type: ignore[method-assign]
        return decorated

    def _decorate_callable(self, decorated: C) -> C:
        if asyncio.iscoroutinefunction(decorated):
            # If the inner function is an async function, we must execute async
            # as well so that the `with` statement executes at the right time.
            @wraps(decorated)
            async def inner(*args: Any, **kwargs: Any) -> Any:
                with self:
                    return await decorated(*args, **kwargs)
        else:

            @wraps(decorated)
            def inner(*args: Any, **kwargs: Any) -> Any:
                with self:
                    return decorated(*args, **kwargs)

        return cast(C, inner)

    def __call__(self, decorated: D) -> D:
        """Decorate TestCase or a callable."""
        if isinstance(decorated, type):
            if issubclass(decorated, TestCase):
                return self._decorate_test_case(decorated)
            else:
                raise TypeError("Can only decorate unittest.TestCase or function.")
        elif callable(decorated):
            return self._decorate_callable(decorated)
        else:
            raise TypeError("Can only decorate unittest.TestCase or function.")
